@@ -306,7 +306,11 @@ PROPERTIES, like DEPLOY-THESE."
 (defprop reconnects :posix (connections properties)
   "Connect back to the same host with CONNECTIONS and apply PROPERTIES.
 Mainly useful for using a connection type like :AS to apply properties as a
-different user."
+different user.
+
+Combinators that work by temporarily pushing hostattrs at :APPLY time will not
+be able to affect PROPERTIES in an application of RECONNECTS they enclose.
+Connection attributes, by contrast, are propagated as usual."
   (:desc (declare (ignore properties))
          (format nil "~S reconnection" connections))
   (:preprocess
@@ -328,10 +332,12 @@ different user."
      (setf (getf (slot-value host 'hostattrs) :data) nil)
      (setq host (preprocess-host host))
      (doplist (k v (hostattrs host))
-              (loop with root = (get-hostattrs k)
-                    for cell on v until (eq cell root)
-                    collect (car cell) into accum
-                    finally (apply #'push-hostattrs k (nreverse accum))))
+       (loop with root = (get-hostattrs k)
+             for cell on v until (eq cell root)
+             collect (car cell) into accum
+             finally (if (eql k :data)
+                         (pushnew-hostattrs :data (nreverse accum))
+                         (apply #'push-hostattrs k (nreverse accum)))))
      (dolist (system (propspec-systems (host-propspec host)))
        (pushnew system (slot-value (host-propspec *host*) 'systems)))
      (setf (getf properties :host) host)))
@@ -346,8 +352,7 @@ different user."
 (defun %propagate-hostattrs (host)
   (dolist (system (propspec-systems (host-propspec host)))
     (pushnew system (slot-value (host-propspec *host*) 'systems)))
-  (dolist (attr (get-hostattrs :data host))
-    (push-hostattrs :data attr)))
+  (pushnew-hostattrs :data (get-hostattrs :data host)))
 
 (defprop evals :posix (&rest forms)
   "Property which just evaluates each of FORMS using EVAL.  Only for testing
